import numpy as np
from pathlib import Path


class DynamicPosition:
    """
    Handler for dynamic positioning files.

    These files contains the results of the mechanical model fits, for each DU in the detector.
    It contains 2 trees:

    - head: This scalar tree contains the generic information about the fit
    - fits: contains the fitted parameters for each DU

    """

    FITS_DTYPE = [
        ("id", np.int32),
        ("tx", np.float64),
        ("ty", np.float64),
        ("tx2", np.float64),
        ("ty2", np.float64),
        ("vs", np.float64),
    ]

    def __init__(self, filename):
        self.filename = Path(filename)
        self._head = None
        self._fits = None
        if self.filename.suffix == ".root":
            self.read_file_root()
        else:
            raise NotImplemented(
                "Error, only dynamic positioning file in root format are supported"
            )

    def read_file_root(self):
        """
        Implement the reading of the root tree so populate self._head and self._fits.
        """
        import uproot

        with uproot.open(self.filename) as fin:
            self._head = fin["ACOUSTICS_FIT/JACOUSTICS::JHead/"].arrays()
            aliases = {
                key.replace("vector<JACOUSTICS::JFit>.", ""): key
                for key in fin["ACOUSTICS_FIT/vector<JACOUSTICS::JFit>/"].keys()
            }
            del aliases["fUniqueID"]
            del aliases["fBits"]
            self._fits = fin["ACOUSTICS_FIT/vector<JACOUSTICS::JFit>/"].arrays(
                aliases.keys(), aliases=aliases
            )

    @property
    def head(self):
        """Return the content of ACOUSTICS_FIT/JACOUSTICS::JHead/"""
        return self._head.to_numpy()

    @property
    def fits(self):
        """Return the content of 'ACOUSTICS_FIT/vector<JACOUSTICS::JFit>/"""
        return self._fits

    def fit(self, DU, fill_value=0):
        """Return the fit for a given DU. Same shape as head, missing fit values are replaced by fill_value"""
        import awkward as ak

        m_fits = self._fits.id == DU
        fits = ak.fill_none(ak.firsts(self._fits[m_fits]), fill_value).to_numpy()
        return fits

    @property
    def DUs(self):
        """Return the list of DUs for which fits are available"""
        import awkward as ak

        return np.unique(ak.flatten(self._fits.id))



import numpy as np
from pathlib import Path
from .detector import Detector


class AuxilliaryObject:

    def __init__(self, filename, detector=None):
        self.filename = Path(filename)
        self._data = dict()
        self._detector = None

        if self.filename.suffix == ".txt":
            with open(self.filename, "r") as fin:
                self._data["raw"] = read_txt_file(fin, self.DTYPE)

        if detector is not None:
            self._set_detector(detector)

    def _set_detector(self, detector):
        """
        Utility class to attach a detector geometry to the Tripods object.
        """
        if isinstance(detector, Detector):
            self._detector = detector

        if isinstance(detector, str):
            detector = Path(detector)

        if isinstance(detector, Path):
            self._detector = Detector(detector)

        self._utm, self._grid = self._detector.utm

        return

    @property
    def detector(self):
        """
        Detector attached to geometry.
        """
        return self._detector

    @property
    def relative(self):
        """
        Tripods coordinates relative to Detector object.
        """
        if self._detector is None:
            raise AttributeError(
                "Detector model not declared, impossible to provide relative coordinates. use set_detector() to associate a detector geometry to the tripods."
            )

        return self._data["relative"]

    @property
    def absolute(self):
        """
        Absolute tripods coordinates.
        """
        return self._data["absolute"]

    @property
    def raw(self):
        """
        Raw (as written in the input file) coordinates.
        """
        return self._data["raw"]


class Tripods(AuxilliaryObject):
    """
    IO class for tripod.txt files, that contains the absolute
    coordinates of acoustic emitters (outside of Base Module emitters,
    that are handled in the Transmitter class.

    A detector can be optionally provided in order to shift the tripod
    in the reference frame of the Detector object.

    Parameters
    ----------
    filename : str or Path
    detector : str or Path or Detector, optional
    """

    DTYPE = np.dtype(
        [
            ("jpp_id", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
        ]
    )

    def __init__(self, filename, detector=None):
        super().__init__(filename, detector)
        self._data["absolute"] = self._data["raw"]
        if detector:
            self._data["relative"] = coordinates_relative_to_reference(
                self._data["absolute"], self._utm
            )


class Transmitters(AuxilliaryObject):
    """
    IO class for transmitter.txt files, that contains the relative
    coordinates of acoustic emitters attached to Base Module. The
    coordinates are expressed w.r.t. to the coordinate a module in the
    detector file, which is refered too by the du:floor pair key, with
    the floor typically being 0.
    A detector can be optionally provided in order to shift the transmitter
    in the reference frame of the Detector object.

    Parameters
    ----------
    filename : str or Path
    detector : str or Path or Detector, optional
    """

    DTYPE = np.dtype(
        [
            ("jpp_id", np.int32),
            ("du", np.int32),
            ("floor", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
        ]
    )

    def __init__(self, filename, detector=None):
        super().__init__(filename, detector)

        if detector:
            self._data["relative"] = generate_coordinates_from_modules_relation(
                self._data["raw"], self._detector.modules
            )
            self._data["absolute"] = coordinates_relative_to_reference(
                self._data["relative"], -1.0 * np.array(self._utm)
            )


class Hydrophones(AuxilliaryObject):
    """
    IO class for hydrophone.txt files, that contains the relative
    coordinates of hydrophones  attached to Base Module. The
    coordinates are expressed w.r.t. to the coordinate a module in the
    detector file, which is refered too by the du:floor pair key, with
    the floor typically being 0.

    A detector can be optionally provided in order to shift the transmitter
    in the reference frame of the Detector object.

    Parameters
    ----------
    filename : str or Path
    detector : str or Path or Detector, optional
    """

    DTYPE = np.dtype(
        [
            ("du", np.int32),
            ("floor", np.int32),
            ("pos_x", np.float64),
            ("pos_y", np.float64),
            ("pos_z", np.float64),
        ]
    )

    def __init__(self, filename, detector=None):
        super().__init__(filename, detector)
        self._data["raw"] = self._data["raw"][self._data["raw"]["floor"] != -1]

        if detector:
            self._data["relative"] = generate_coordinates_from_modules_relation(
                self._data["raw"], self._detector.modules
            )
            self._data["absolute"] = coordinates_relative_to_reference(
                self._data["relative"], -1.0 * np.array(self._utm)
            )


def read_txt_file(io, dtype):
    """
    IO class to read the txt file and return a Nammed Numpy array.
    """
    lines = io.readlines()

    lines = [
        line.strip() for line in lines if line.strip() and not line.startswith("#")
    ]

    data = []

    for line in lines:
        elements = line.split()
        data.append(tuple(elements))

    return np.array(data, dtype=dtype)


def coordinates_relative_to_reference(data, reference):
    """
    Compute the coordinate in the relative to reference.
    """
    relative = data.copy()

    relative["pos_x"] -= reference[0]
    relative["pos_y"] -= reference[1]
    relative["pos_z"] -= reference[2]

    return relative


def generate_coordinates_from_modules_relation(data, modules):
    """
    Compute the coordinate shifted with respect to the (du,floor) key indicated.
    """

    relative = data.copy()

    shifts = []

    for row in data:
        du, floor = row["du"], row["floor"]

        mod = modules[(modules["du"] == du) & (modules["floor"] == floor)]
        if len(mod) == 0:
            raise ValueError(
                f"Transmitter is linked to DU{du:03d}F{floor:02d}, but this module doesn't exist in the attached detector"
            )

        mod = mod[0]

        shifts.append([mod["pos_x"], mod["pos_y"], mod["pos_z"]])
    shifts = np.stack(shifts)

    relative["pos_x"] += shifts[:, 0]
    relative["pos_y"] += shifts[:, 1]
    relative["pos_z"] += shifts[:, 2]

    return relative



import numpy as np
from numpy.lib import recfunctions as rfn
from pathlib import Path
from .detector import Detector


class DynamicOrientation:
    """
    Handler for dynamic orientation files.
    """

    def __init__(self, filename, detector=None):
        self.filename = Path(filename)
        self._fits = None
        if self.filename.suffix == ".root":
            self.read_file_root()
        else:
            raise NotImplemented(
                "Error, only dynamic  file in root format are supported"
            )

        if detector:
            self._detector = Detector(detector)

    def read_file_root(self):
        """
        Implement the reading of the root tree so populate self._head and self._fits.
        """
        import uproot

        with uproot.open(self.filename) as fin:
            tree = fin["ORIENTATION/ORIENTATION"]
            keys = ["id", "t", "a", "b", "c", "d", "roll", "pitch", "yaw"]
            aliases = {
                "roll": "atan2(2 * (a*b + c*d), 1 - 2 * (b**2 + c**2))",
                "pitch": "arcsin(2 * (a*c - d*b))",
                "yaw": "atan2(2 * (a*d + b*c), 1 - 2 * (b**2 + d**2))",
            }
            self._fits = tree.arrays(keys, aliases=aliases).to_numpy()

    @property
    def du_floor(self):
        """
        Enrich the fits informations with the du:floor position from detector file
        """
        mapping = self._detector.module_id_mapping
        du_floor = np.stack([mapping[module] for module in self._fits["id"]])
        return du_floor[:, 0], du_floor[:, 1]

    @property
    def fits(self):
        """Return the content of 'ORIENTATION/"""
        return self._fits



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
from km3net_testdata import data_path
from km3det import Detector, Tripods, Hydrophones, Transmitters


class TestDetectorV5(unittest.TestCase):
    def setUp(self):
        self.datx_path = data_path("detector_geometry/KM3NeT_00000232.detector.datx")
        self.tripod_path = data_path("detector_geometry/KM3NeT_00000232.tripod.txt")
        self.hydrophone_path = data_path(
            "detector_geometry/KM3NeT_00000232.hydrophone.txt"
        )
        self.transmitter_path = data_path(
            "detector_geometry/KM3NeT_00000232.transmitter.txt"
        )
        self.detector = Detector(self.datx_path)

    def test_tripod_read(self):
        tripod = Tripods(self.tripod_path)
        tripod = Tripods(self.tripod_path, self.datx_path)
        tripod = Tripods(self.tripod_path, self.detector)

        assert len(tripod.raw) == 8
        assert len(tripod.absolute) == 8
        assert len(tripod.relative) == 8

    def test_hydrophone_read(self):
        hydrophone = Hydrophones(self.hydrophone_path)
        hydrophone = Hydrophones(self.hydrophone_path, self.datx_path)
        hydrophone = Hydrophones(self.hydrophone_path, self.detector)

        assert len(hydrophone.raw) == 29
        assert len(hydrophone.absolute) == 29
        assert len(hydrophone.relative) == 29

    def test_transmitter_read(self):
        transmitter = Transmitters(self.transmitter_path)
        transmitter = Transmitters(self.transmitter_path, self.datx_path)
        transmitter = Transmitters(self.transmitter_path, self.detector)

        assert len(transmitter.raw) == 5
        assert len(transmitter.absolute) == 5
        assert len(transmitter.relative) == 5
